import polars as pl


def downcast_strings(df: pl.DataFrame | pl.LazyFrame) -> pl.DataFrame | pl.LazyFrame:
    original_schema = df.schema


    transform_cols = [col for col in df.columns if original_schema[col] == pl.Utf8]
    if len(transform_cols) == 0:
        return df
    
    transformed_df = df.with_columns([pl.col(col).cast(pl.Categorical) for col in df.columns if original_schema[col] == pl.Utf8])

    return transformed_df
            

def test_downcast_strings_eager():
    df = pl.DataFrame({
        "a": ["a", "b", "c"],
        "b": [1, 2, 3],
        "c": [True, False, True]
    })

    assert df.schema["a"] == pl.Utf8
    assert df.schema["b"] == pl.Int64
    assert df.schema["c"] == pl.Boolean

    df = downcast_strings(df)

    assert df.schema["a"] == pl.Categorical, f"Expected {df.schema['a']} to be {pl.Categorical}"
    assert df.schema["b"] == pl.Int64,f"Expected {df.schema['b']} to be {pl.Int64}"
    assert df.schema["c"] == pl.Boolean,f"Expected {df.schema['c']} to be {pl.Boolean}"

def test_downcast_strings_lazy():
    df = pl.DataFrame({
        "a": ["a", "b", "c"],
        "b": [1, 2, 3],
        "c": [True, False, True]
    }).lazy()

    assert df.schema["a"] == pl.Utf8
    assert df.schema["b"] == pl.Int64
    assert df.schema["c"] == pl.Boolean

    df = downcast_strings(df)

    df = df.collect()
    assert df.schema["a"] == pl.Categorical, f"Expected {df.schema['a']} to be {pl.Categorical}"
    assert df.schema["b"] == pl.Int64,f"Expected {df.schema['b']} to be {pl.Int64}"
    assert df.schema["c"] == pl.Boolean,f"Expected {df.schema['c']} to be {pl.Boolean}"



import polars as pl
from .integers import downcast_integers
from .strings import downcast_strings

def compress_frame(df):
    df = downcast_strings(df)
    df = downcast_integers(df)
    return df

def real_test_compresss_file():
    df = pl.scan_csv("exampledata1.csv", infer_schema_length=1)
    print(f"Original schema: {df.schema}")
    df = compresss_file(df).collect()
    print(f"Schema: {df.schema}")



import polars as pl

def is_attribute(col, df: pl.LazyFrame):
    return df.select(pl.col(col).unique()).collect().n_unique() == 1

def test_is_attribute():
    test_data = pl.DataFrame({
        'a' : [1, 2, 3],
        'b' : [1, 1, 1] }).lazy()
    assert is_attribute('a', test_data) == False
    assert is_attribute('b', test_data) == True
    print("is_attribute() passed")

test_is_attribute()

def extract_attributes(df: pl.LazyFrame | pl.DataFrame) -> pl.DataFrame | pl.LazyFrame:   
    start_mode = 'lazy' if isinstance(df, pl.LazyFrame) else 'eager'
    df = df.lazy() if start_mode == 'eager' else df
    static_cols = [col for col in df.columns if is_attribute(col, df)]
    attribute_frame = df.select(static_cols)
    fact_frame = df.select([col for col in df.columns if col not in static_cols])
    attribute_frame = attribute_frame.unique()
    if start_mode == "lazy":
        return attribute_frame, fact_frame
    return attribute_frame.collect(), fact_frame.collect()

def test_extract_attributes():
    test_data = pl.DataFrame({
        'a' : [1, 2, 3],
        'b' : [1, 1, 1] }).lazy()
    attribute_frame, fact_frame = extract_attributes(test_data)
    assert attribute_frame.collect().frame_equal(pl.DataFrame({'b' : [1]})), attribute_frame.collect()
    assert fact_frame.collect().frame_equal(pl.DataFrame({'a' : [1, 2, 3]})), fact_frame.collect()
    print("extract_attributes() passed for lazy frames")
    test_data = test_data.collect()
    attribute_frame, fact_frame = extract_attributes(test_data)
    assert attribute_frame.frame_equal(pl.DataFrame({'b' : [1]})), attribute_frame
    assert fact_frame.frame_equal(pl.DataFrame({'a' : [1, 2, 3]})), fact_frame
    print("extract_attributes() passed for eager frames")

test_extract_attributes()
    
    



# this module is focused on identifying and extracting static columns

import polars as pl

def identify_static_cols(df: pl.DataFrame | pl.LazyFrame)-> list[str]:
    """
    Identify columns that are static attributes
    """
    # loop through df.columns and identify columns that are static
    df =df.lazy()
    return [col for col in df.columns if df.select(pl.col(col).n_unique()).collect()[col].to_list()[0]==1]    

def extract_static_cols(df: pl.DataFrame | pl.LazyFrame)-> dict[str, pl.DataFrame | pl.LazyFrame ]:
    """
    Remove columns that are static attributes
    """
    static_cols = identify_static_cols(df)
    attr_df = df.select(static_cols)
    attributes = attr_df.select(pl.all().unique()) #get the unique values of each column
    if not isinstance(attr_df, pl.DataFrame):
        attributes = attr_df.collect()
    
    # get the static values of each of the attr_df columns
    #use select to get all the values of each column
    data = df.select([col for col in df.columns if col not in static_cols])
    
    return {"data": data, "attributes": attributes}


#TESTS
  
def test_identify_static_cols():
    test_frame = pl.DataFrame({
        "a": [1,1,1,1,1,1,1,1,1,1], 
        "b": [1,1,1,1,1,1,1,1,1,1],
        "c": [1,1,1,1,1,1,1,1,1,1],
        'x': [1,2,3,4,5,6,7,8,9,10],
        #string columns
        "d": ["a","a","a","a","a","a","a","a","a","a"],
        "e": ["a","a","a","a","a","a","a","a","a","a"],
        "ecat": ["a","a","a","a","a","a","a","a","a","a"],
        #float columns
        "f": [1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
        "g": [1.0,0.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
        #bool columns
        'h': [True, True, True, True, True, True, True, True, True, True],
        'i': [True, True, True, True, True, True, True, True, True, True],
        'j': [True, True, True, False, True, True, True, True, True, True]
    })
    # Convert one string column to categorical
    test_frame = test_frame.with_columns(pl.col("ecat").cast(pl.Categorical))
    #Test function
    test_values = ('a', 'b', 'c', 'd', 'e', 'ecat', 'f', 'h', 'i')
    static_cols = identify_static_cols(test_frame)
    missing = {col for col in test_values if col not in static_cols}
    added = {col for col in static_cols if col not in test_values}
    assert tuple(static_cols) == test_values, f"static_cols missing{missing} added{added}"
    #Test function with lazy frame'
    test_frame = test_frame.lazy()
    static_cols = identify_static_cols(test_frame)
    missing = {col for col in test_values if col not in static_cols}
    added = {col for col in static_cols if col not in test_values}
    assert tuple(static_cols) == test_values, f"static_cols missing{missing} added{added}"
    print("Test passed")
        
#make call to test func
def test_extract_static_cols():
    test_frame = pl.DataFrame({
        "a": [1,1,1,1,1,1,1,1,1,1], 
        "b": [1,1,1,1,1,1,1,1,1,1],
        "c": [1,1,1,1,1,1,1,1,1,1],
        'x': [1,2,3,4,5,6,7,8,9,10],
        #string columns
        "d": ["a","a","a","a","a","a","a","a","a","a"],
        "e": ["a","a","a","a","a","a","a","a","a","a"],
        "ecat": ["a","a","a","a","a","a","a","a","a","a"],
        #float columns
        "f": [1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
        "g": [1.0,0.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
        #bool columns
        'h': [True, True, True, True, True, True, True, True, True, True],
        'i': [True, True, True, True, True, True, True, True, True, True],
        'j': [True, True, True, False, True, True, True, True, True, True]
    })
    # Convert one string column to categorical
    test_frame = test_frame.with_columns(pl.col("ecat").cast(pl.Categorical))
    #Test function
    test_values = ('a', 'b', 'c', 'd', 'e', 'ecat', 'f', 'h', 'i')
    static_cols = extract_static_cols(test_frame)
    comparison_attributes = pl.DataFrame({
        "a": [1],
        "b": [1],
        "c": [1],
        "d": ["a"],
        "e": ["a"],
        "ecat": ["a"],
        "f": [1.0],
        "h": [True],
        "i": [True]
    })
    
    comparison_data = test_frame.select([col for col in test_frame.columns if col not in test_values])
    
    assert static_cols["attributes"].frame_equal(comparison_attributes), f"attributes not equal"
    assert static_cols["data"].frame_equal(comparison_data), f"data not equal"
    print("Test passed")



import polars as pl
from integers import downcast_integers
from strings import downcast_strings
from attributes import extract_attributes
from relaxed_concat import majority_schema_concat
from typing import Optional
import os

class ComplexFrame:
    """
    This class defines the complex frame, which is a collection of
    all internal frames are primarily related or come from the same source frame 
    """
    io_dict = {'input':
        {'csv': pl.read_csv, 'parquet': pl.read_parquet, 'ipc': pl.read_ipc, 'json': pl.read_json},
        'output':
        {'csv': pl.DataFrame.to_csv, 'parquet': pl.DataFrame.to_parquet, 'ipc': pl.DataFrame.to_ipc, 'json': pl.DataFrame.to_json}
    }
    
    def __init__(self, frame : pl.LazyFrame | pl.LazyFrame, attributes: Optional[pl.DataFrame | dict] = None):
        self.frames = {}
        compressed_frame = downcast_integers(frame)
        categorized_frame = downcast_strings(compressed_frame)
        self.frames['attributes'], self.frames['facts'] = extract_attributes(categorized_frame)
        if attributes is not None:
            self.frames['attributes'] = attributes
    
    def to_lazy(self, frame):
        """
        Convert a complex frame to a lazy frame
        """
        #left join attributes to facts using cross join
        return self.frames['facts'].join(
            self.frames['attributes'], on = None, how = 'cross')
    @staticmethod
    def write_directory(self, directory_path: str, how : str = 'parquet' | 'csv' | 'json' | 'ipc' ):
        """
        Write a complex frame to a directory
        """
        directory_info = []
        for name, frame in self.frames.items():
            directory_info.append(f"{directory_path}\\{name}.{how}")
            self.io_dict['output'][how](frame, directory_info[-1])
        return directory_info
    
    @classmethod
    def read_directory(cls, directory_path: str):
        """
        Read a complex frame from a directory
        """
        frames = {}
        for filepath in os.listdir(directory_path):
            name = filepath.split('.')[-1]
            how = filepath.split('.')[-2]
            frames[name] = cls.io_dict['input'][how](f"{directory_path}\\{filepath}")
        return cls(frames['attributes'], frames['facts'])
            
        


def complex_concat(frames: iter[ComplexFrame]):
    """
    Concatenate a list of frames into a single frame
    """
    lazy_frames = [frame.to_lazy(frame) for frame in frames]
    lazy_frame = majority_schema_concat(lazy_frames)
    return ComplexFrame(lazy_frame)

    
        
    
        
        
        
        
        